'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

const indices = ["First", "Second", "Third"];
function validateArguments(method, options) {
    options.map(([value, expected, required], index) => {
        const expectedName = typeof expected === "string" ? expected : expected.name;
        const matches = typeMatches(expected, value);
        return invariant(required ? matches : value === undefined || value === null || matches, `${indices[index]} argument of "${method}" should be of type: "${expectedName}", received: "${typeof value}".`);
    });
}
function typeMatches(expected, value) {
    if (typeof expected === "string") {
        const type = Array.isArray(value) ? "array" : typeof value;
        return type === expected;
    }
    return value instanceof expected;
}
function invariant(condition, message) {
    if (!condition) {
        throw new Error(message);
    }
}

function raw(value) {
    return new RawArgument(value);
}
class RawArgument {
    constructor(value) {
        this.value = value;
    }
}

function token(name) {
    return new Token(name);
}
class Token {
    constructor(name) {
        this.type = null;
        this.name = name;
    }
}

class Providers {
    constructor() {
        this.providers = [];
    }
    add(tokens, callback, called = false) {
        tokens.forEach((token) => {
            invariant(token instanceof Token, `Trying to provide non token argument.`);
            invariant(!this.has(token), `Trying to provide dependency "${token.name}" which is already provided.`);
        });
        this.providers.push({ tokens, callback, called });
    }
    has(token) {
        return Boolean(this.get(token));
    }
    notify(token) {
        const provider = this.get(token);
        if (provider) {
            provider.callback();
            provider.called = true;
        }
    }
    get(token) {
        return this.providers.find((provider) => !provider.called &&
            provider.tokens.some((providerToken) => providerToken.name === token.name));
    }
}

class Requests {
    constructor() {
        this.requests = [];
    }
    add(token, callback) {
        this.requests.push({ token, callback });
    }
    fulfill(token) {
        this.requests
            .filter((request) => request.token.name === token.name)
            .forEach((request) => request.callback());
        this.remove(token);
    }
    has(token) {
        return this.requests.some((request) => request.token.name === token.name);
    }
    remove(token) {
        this.requests = this.requests.filter((request) => request.token.name !== token.name);
    }
}

class Dependencies {
    constructor() {
        this.dependencies = new Map();
    }
    add(token, creator, args = [], shared = false) {
        invariant(!this.dependencies.get(token.name), `Dependency "${token.name}" already exists.`);
        invariant(!args.some((arg) => arg instanceof Token && arg.name === token.name), `Trying to add a recursive dependency "${token.name}".`);
        this.dependencies.set(token.name, { token, creator, args, shared });
    }
    get(token) {
        invariant(this.has(token), `Dependency "${token.name}" does not exist.`);
        return this.dependencies.get(token.name);
    }
    has(token) {
        return this.dependencies.has(token.name);
    }
}

class Container {
    constructor() {
        this.providers = new Providers();
        this.requests = new Requests();
        this.dependencies = new Dependencies();
    }
    add(token, creator, args = []) {
        this.register(token, creator, args, "add");
    }
    share(token, creator, args = []) {
        this.register(token, creator, args, "share", true);
    }
    constant(token, value) {
        this.register(token, () => value, [], "constant", true);
    }
    register(token, creator, args, method, shared = false) {
        validateArguments(method, [
            [token, Token, true],
            [creator, "function", true],
            [args, "array"],
        ]);
        this.dependencies.add(token, creator, args, shared);
        this.requests.fulfill(token);
    }
    get(token) {
        validateArguments("get", [[token, Token, true]]);
        if (this.providers.has(token)) {
            this.providers.notify(token);
        }
        return this.instantiate(this.dependencies.get(token));
    }
    getAsync(token) {
        validateArguments("getAsync", [[token, Token, true]]);
        if (this.providers.has(token)) {
            this.providers.notify(token);
        }
        if (this.dependencies.has(token)) {
            return Promise.resolve(this.instantiate(this.dependencies.get(token)));
        }
        return new Promise((resolve) => {
            this.requests.add(token, () => {
                this.instantiateAsync(this.dependencies.get(token), resolve);
            });
        });
    }
    provide(tokens) {
        validateArguments("provide", [[tokens, "array", true]]);
        const existing = tokens.find((token) => this.dependencies.has(token));
        invariant(!existing, `Trying to provide dependency "${existing && existing.name}" which already exists.`);
        if (tokens.some((token) => this.requests.has(token))) {
            this.providers.add(tokens, () => { }, true);
            return Promise.resolve();
        }
        let resolver;
        const promise = new Promise((resolve) => {
            resolver = resolve;
        });
        this.providers.add(tokens, resolver);
        return promise;
    }
    instantiate(dependency) {
        if (dependency.shared) {
            return dependency.instance || this.createSingle(dependency);
        }
        return this.create(dependency);
    }
    instantiateAsync(dependency, callback) {
        if (dependency.shared) {
            return dependency.instance
                ? callback(dependency.instance)
                : this.createSingleAsync(dependency, callback);
        }
        return this.createAsync(dependency, callback);
    }
    create(dependency) {
        return this.construct(dependency.creator, this.resolve(dependency.args));
    }
    createAsync(dependency, callback) {
        this.resolveAsync(dependency.args, (resolvedArgs) => callback(this.construct(dependency.creator, resolvedArgs)));
    }
    createSingle(dependency) {
        dependency.instance = this.create(dependency);
        return dependency.instance;
    }
    createSingleAsync(dependency, callback) {
        this.createAsync(dependency, (newInstance) => {
            dependency.instance = newInstance;
            callback(newInstance);
        });
    }
    resolve(args) {
        return args.map((arg) => arg instanceof RawArgument ? arg.value : this.get(arg));
    }
    resolveAsync(args, callback) {
        args.reduceRight((next, arg) => (resolvedArgs) => {
            return arg instanceof RawArgument
                ? next([...resolvedArgs, arg.value])
                : this.getAsync(arg).then((dependency) => next([...resolvedArgs, dependency]));
        }, callback)([]);
    }
    construct(creator, args) {
        return creator.hasOwnProperty("prototype")
            ? new creator(...args)
            : creator(...args);
    }
}
function createContainer() {
    return new Container();
}

exports.Container = Container;
exports.RawArgument = RawArgument;
exports.Token = Token;
exports.createContainer = createContainer;
exports.raw = raw;
exports.token = token;
